#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016

import collections
import itertools
import io

from ..mkExceptions import BadPubmedRecord, RCTypeError
from ..mkRecord import ExtendedRecord
from .tagProcessing.tagNames import tagNameConverterDict, authorBasedTags
from .tagProcessing.tagFunctions import medlineTagToFunc
from .tagProcessing.specialFunctions import medlineSpecialTagToFunc

class MedlineRecord(ExtendedRecord):
    """Class for full Medline(Pubmed) entries.

    This class is an [ExtendedRecord](./ExtendedRecord.html#metaknowledge.ExtendedRecord) capable of generating its own id number. You should not create them directly, but instead use [medlineParser()](../modules/medline.html#metaknowledge.medline.medlineHandlers.medlineParser) on a medline file.
    """
    def __init__(self, inRecord, sFile = "", sLine = 0):
        bad = False
        error = None
        fieldDict = None
        try:
            if isinstance(inRecord, dict) or isinstance(inRecord, collections.OrderedDict):
                fieldDict = collections.OrderedDict(inRecord)
            elif isinstance(inRecord, itertools.chain):
                fieldDict = medlineRecordParser(inRecord)
            elif isinstance(inRecord, io.IOBase):
                fieldDict = medlineRecordParser(enumerate(inRecord))
            elif isinstance(inRecord, str):
                def addCharToEnd(lst):
                    for s in lst:
                        yield s + '\n'
                fieldDict = medlineRecordParser(enumerate(addCharToEnd(inRecord.split('\n')), start = 1))
                #string io
            else:
                raise RCTypeError("Unsupported input type '{}', PubmedRecords cannot be created from '{}'".format(inRecord, type(inRecord)))
        except BadPubmedRecord as b:
            self.bad = True
            self.error = b
            fieldDict = collections.OrderedDict()
        if fieldDict is not None:
            if 'PMID' in fieldDict:
                self._pubNum = "PMID:{}".format(fieldDict['PMID'][0])
            else:
                self._pubNum = None
                bad = True
                error = BadPubmedRecord("Missing PMID")
        ExtendedRecord.__init__(self, fieldDict, self._pubNum, bad, error, sFile = sFile, sLine = sLine)

    def encoding(self):
        return 'latin-1'

    @staticmethod
    def getAltName(tag):
        return tagNameConverterDict.get(tag)

    @staticmethod
    def tagProcessingFunc(tag):
        return medlineTagToFunc[tag]

    def specialFuncs(self, key):
        #This will usually raise a key error that needs to be caught higher up
        #Catching it here would require unnecessary overhead
        return medlineSpecialTagToFunc[key](self)

    def writeRecord(self, f):
        """This is nearly identical to the original the FAU tag is the only tag not writen in the same place, doing so would require changing the parser and lots of extra logic.
        """
        if self.bad:
            raise BadPubmedRecord("This record cannot be converted to a file as the input was malformed.\nThe original line number (if any) is: {} and the original file is: '{}'".format(self._sourceLine, self._sourceFile))
        else:
            authTags = {}
            for tag in authorBasedTags:
                for val in self._fieldDict.get(tag, []):
                    split = val.split(' : ')
                    try:
                        authTags[split[0]].append("{0}{1}- {2}\n".format(tag, ' ' * (4 - len(tag)),' : '.join(split[1:]).replace('\n', '\n      ')))
                    except KeyError:
                        authTags[split[0]] = ["{0}{1}- {2}\n".format(tag, ' ' * (4 - len(tag)),' : '.join(split[1:]).replace('\n', '\n      '))]
            for tag, value in self._fieldDict.items():
                if tag in authorBasedTags:
                    continue
                else:
                    for v in value:
                        f.write("{0}{1}- {2}\n".format(tag, ' ' * (4 - len(tag)), v.replace('\n', '\n      ')))
                        if tag == 'AU':
                            for authVal in authTags.get(v,[]):
                                f.write(authVal)

def medlineRecordParser(record):
    """The parser [`MedlineRecord`](../classes/MedlineRecord.html#metaknowledge.medline.MedlineRecord) use. This takes an entry from [medlineParser()](#metaknowledge.medline.medlineHandlers.medlineParser) and parses it a part of the creation of a `MedlineRecord`.

    # Parameters

    _record_ : `enumerate object`

    > a file wrapped by `enumerate()`

    # Returns

    `collections.OrderedDict`

    > An ordered dictionary of the key-vaue pairs in the entry
    """
    tagDict = collections.OrderedDict()
    tag = 'PMID'
    mostRecentAuthor = None
    for lineNum, line in record:
        tmptag = line[:4].rstrip()
        contents = line[6:-1]
        if tmptag.isalpha() and line[4] == '-':
            tag = tmptag
            if tag == 'AU':
                mostRecentAuthor = contents
            if tag in authorBasedTags:
                contents = "{} : {}".format(mostRecentAuthor, contents)
            try:
                tagDict[tag].append(contents)
            except KeyError:
                tagDict[tag] = [contents]
        elif line[:6] == '      ':
            tagDict[tag][-1] += '\n' + line[6:-1]
        elif line == '\n':
            break
        else:
            raise BadPubmedRecord("Tag not formed correctly on line {}: '{}'".format(lineNum, line))
    return tagDict



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
from .helpFuncs import getMonth
from ...citation import Citation

import collections

def pubType(val):
    """
    # The PT Tag

    extracts the type of publication as a character: conference, book, journal, book in series, or patent

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > A string

    """
    return val[0]


def authorsFull(val):
    """
    # The AF Tag

    extracts a list of authors full names

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > A list of author's names

    """
    return val

def group(val):
    """
    # The GP Tag

    extracts the group associated with the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > A the name of the group

    """
    return val[0]

def editedBy(val):
    """
    # The BE Tag

    extracts a list of the editors of the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > A list of editors

    """
    return val

def authorsShort(val):
    """
    # The AU Tag

    extracts a list of authors shortened names

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > A list of shortened author's names

    """
    return val

def bookAuthor(val):
    """
    # The BA Tag

    extracts a list of the short names of the authors of a book Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > A list of shortened author's names

    """
    return val

def bookAuthorFull(val):
    """
    # The BF Tag

    extracts a list of the long names of the authors of a book Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > A list of author's names

    """
    return val

def groupName(val):
    """
    # The CA Tag

    extracts the name of the group associated with the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The group's name

    """
    return val[0]

def title(val):
    """
    # The TI Tag

    extracts the title of the record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The title of the record

    """
    return ' '.join(val)

def editors(val):
    """
    # Needs Work

    currently not well understood, returns _val_
    """
    return val

def journal(val):
    """
    # The SO Tag

    extracts the full name of the publication and normalizes it to uppercase

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The name of the journal

    """
    return ' '.join(val).upper()

def seriesTitle(val):
    """
    # The SE Tag

    extracts the title of the series the Record is in

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The title of the series

    """
    return ' '.join(val)

def seriesSubtitle(val):
    """
    # The BS Tag

    extracts the title of the series the Record is in

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The subtitle of the series

    """
    return ' '.join(val)

def language(val):
    """
    # The LA Tag

    extracts the languages of the Record as a string with languages separated by ', ', usually there is only one language

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The language(s) of the record

    """
    return ', '.join(val)

def docType(val):
    """
    # The DT Tag

    extracts the type of document the Record contains

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The type of the Record

    """
    return val[0]

def confTitle(val):
    """
    # The CT Tag

    extracts the title of the conference associated with the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The title of the conference

    """
    return ' '.join(val)

def confDate(val):
    """
    # The CY Tag

    extracts the date string of the conference associated with the Record, the date is not normalized

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The data of the conference

    """
    return val[0]

def confSponsors(val):
    """
    # The SP Tag

    extracts a list of sponsors for the conference associated with the record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > A the list of of sponsors

    """
    return ', '.join(val).split(', ')

def wosTimesCited(val):
    """
    # The TC Tag

    extracts the number of times the Record has been cited by records in WOS

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `int`

    > The number of time the Record has been cited

    """
    return int(val[0])

def authAddress(val):
    """
    # The C1 Tag

    extracts the address of the authors as given by WOS. **Warning** the mapping of author to address is not very good and is given in multiple ways.

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > A list of addresses

    """
    ret = []
    for a in val:
        if a[0] == '[':
            ret.append('] '.join(a.split('] ')[1:]))
        else:
            ret.append(a)
    return ret

def confLocation(val):
    """
    # The CL Tag

    extracts the sting giving the conference's location

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The conferences address

    """
    return ' '.join(val)

def j9(val):
    """
    # The J9 Tag

    extracts the J9 (29-Character Source Abbreviation) of the publication

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The 29-Character Source Abbreviation

    """
    return val[0]

def funding(val):
    """
    # The FU Tag

    extracts a list of the groups funding the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > A list of funding groups

    """
    return ' '.join(val).split('; ')

def subjectCategory(val):
    """
    # The SC Tag

    extracts a list of the subjects associated with the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > A list of the subjects associated with the Record

    """
    return ' '.join(val).split('; ')

def citations(val):
    """
    # The CR Tag

    extracts a list of all the citations in the record, the citations are the [metaknowledge.Citation](../classes/Citation.html#metaknowledge.citation.Citation) class.

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    ` list[metaknowledge.Citation]`

    > A list of Citations

    """
    retCites = []
    for c in val:
        retCites.append(Citation(c))
    return retCites

def publisherCity(val):
    """
    # The PI Tag

    extracts the city the publisher is in

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The city of the publisher

    """
    return ' '.join(val).upper()

def ISSN(val):
    """
    # The SN Tag

    extracts the ISSN of the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The ISSN string

    """
    return val[0]

def articleNumber(val):
    """
    # The AR Tag

    extracts a string giving the article number, not all are integers

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The article number

    """
    return val[0]

def issue(val):
    """
    # The IS Tag

    extracts a string giving the issue or range of issues the Record was in, not all are integers

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The issue number/range

    """
    return val[0]

def email(val):
    """
    # The EM Tag

    extracts a list of emails given by the authors of the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > A list of emails

    """
    return ' '.join(val).split('; ')

def eISSN(val):
    """
    # The EI Tag

    extracts the EISSN of the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The EISSN string

    """
    return val[0]

def DOI(val):
    """
    # The DI Tag

    return the DOI number of the record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The DOI number string

    """
    return val[0]

def wosString(val):
    """
    # The UT Tag

    extracts the WOS number of the record as a string preceded by "WOS:"

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The WOS number

    """
    return val[0]

def orcID(val):
    """
    # The OI Tag

    extracts a list of orc IDs of the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The orc ID

    """
    return ' '.join(val).split('; ')

def meetingAbstract(val):
    """
    # The MA Tag

    extracts the ID of the meeting abstract prefixed by 'EPA-'

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The meeting abstract prefixed

    """
    return val[0]

def isoAbbreviation(val):
    """
    # The JI Tag

    extracts the iso abbreviation of the journal

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The iso abbreviation of the journal

    """
    return ' '.join(val).upper()

def pageCount(val):
    """
    # The PG Tag

    returns an integer giving the number of pages of the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `int`

    > The page count

    """
    return int(val[0])

def publisher(val):
    """
    # The PU Tag

    extracts the publisher of the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The publisher

    """
    return ' '.join(val).upper()

def ISBN(val):
    """
    # The BN Tag

    extracts a list of ISBNs associated with the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list`

    > The ISBNs

    """
    return ' '.join(val).split('; ')

def month(val):
    """
    # The PD Tag

    extracts the month the record was published in as an int with January as 1, February 2, ...

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `int`

    > A integer giving the month

    """
    return getMonth(val[0].split(' ')[0])

def fundingText(val):
    """
    # The FX Tag

    extracts a string of the funding thanks

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The funding thank-you

    """
    return ' '.join(val)

def bookDOI(val):
    """
    # The D2 Tag

    extracts the book DOI of the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The DOI number

    """
    return val[0]

def volume(val):
    """
    # The VL Tag

    return the volume the record is in as a string, not all are integers

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The volume number

    """
    return val[0].strip()

def ResearcherIDnumber(val):
    """
    # The RI Tag

    extracts a list of the research IDs of the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > The list of the research IDs

    """
    return ' '.join(val).split('; ')

def citedRefsCount(val):
    """
    # The NR Tag

    extracts the number citations, length of CR list

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `int`

    > The number of CRs

    """
    return int(val[0])

def beginningPage(val):
    """
    # The BP Tag

    extracts the first page the record occurs on, not all are integers

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The first page number

    """
    return val[0].strip()

def abstract(val):
    """
    # The AB Tag

    return abstract of the record, with newlines hopefully in the correct places

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The abstract

    """
    return '\n'.join(val)

def supplement(val):
    """
    # The SU Tag

    extracts the supplement number

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The supplement number

    """
    return ' '.join(val)

def confHost(val):
    """
    # The HO Tag

    extracts the host of the conference

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The host

    """
    return ' '.join(val)

def publisherAddress(val):
    """
    # The PA Tag

    extracts the publishers address

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The publisher address

    """
    return ' '.join(val)

def endingPage(val):
    """
    # The EP Tag

    return the last page the record occurs on as a string, not aall are intergers

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The final page number

    """
    return val[0].strip()

def year(val):
    """
    # The PY Tag

    extracts the year the record was published in as an int

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `int`

    > The year

    """
    yearField = val[0]
    if len(yearField) == 4:
        return int(yearField)
    else:
        raise ValueError("Incorrectly formatted PY tag")

def authKeywords(val):
    """
    # The DE Tag

    extracts the keywords assigned by the author of the Record. The WOS description is:

        Author keywords are included in records of articles from 1991 forward. They are also include in conference proceedings records.

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > The list of keywords

    """
    return ' '.join(val).lower().split('; ')

def reprintAddress(val):
    """
    # The RP Tag

    extracts the reprint address string

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The reprint address

    """
    return val[0]

def totalTimesCited(val):
    """
    # The Z9 Tag

    extracts the total number of citations of the record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `int`

    > The total number of citations

    """
    return int(val[0])

def partNumber(val):
    """
    # The PN Tag

    return an integer giving the part of the issue the Record is in

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `int`

    > The part of the issue of the Record

    """
    return val[0]

def specialIssue(val):
    """
    # The SI Tag

    extracts the special issue value

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The special issue value

    """
    return val[0]

def subjects(val):
    """
    # The WC Tag

    extracts a list of subjects as assigned by WOS

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > The subjects list

    """
    return ' '.join(val).split('; ')

def keywords(val):
    """
    # The ID Tag

    extracts the WOS keywords of the Record. The WOS description is:

        KeyWords Plus are index terms created by Thomson Reuters from significant, frequently occurring words in the titles of an article's cited references.

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `list[str]`

    > The keyWords list

    """
    return ' '.join(val).upper().split('; ')

def pubMedID(val):
    """
    # The PM Tag

    extracts the pubmed ID of the record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The pubmed ID

    """
    return val[0]

def documentDeliveryNumber(val):
    """
    # The GA Tag

    extracts the document delivery number of the Record

    # Parameters

    _val_: `list[str]`

    > The raw data from a WOS file

    # Returns

    `str`

    > The document delivery number

    """
    return val[0]

tagToFunc = collections.OrderedDict([
            ('PT', pubType),
            ('AF', authorsFull),
            ('GP', group),
            ('BE', editedBy),
            ('AU', authorsShort),
            ('BA', bookAuthor),
            ('BF', bookAuthorFull),
            ('CA', groupName),
            ('ED', editors),
            ('TI', title),
            ('SO', journal),
            ('SE', seriesTitle),
            ('BS', seriesSubtitle),
            ('LA', language),
            ('DT', docType),
            ('CT', confTitle),
            ('CY', confDate),
            ('HO', confHost),
            ('CL', confLocation),
            ('SP', confSponsors),
            ('DE', authKeywords),
            ('ID', keywords),
            ('AB', abstract),
            ('C1', authAddress),
            ('RP', reprintAddress),
            ('EM', email),
            ('RI', ResearcherIDnumber),
            ('OI', orcID),
            ('FU', funding),
            ('FX', fundingText),
            ('CR', citations),
            ('NR', citedRefsCount),
            ('TC', wosTimesCited),
            ('Z9', totalTimesCited),
            ('PU', publisher),
            ('PI', publisherCity),
            ('PA', publisherAddress),
            ('SC', subjectCategory),
            ('SN', ISSN),
            ('EI', eISSN),
            ('BN', ISBN),
            ('J9', j9),
            ('JI', isoAbbreviation),
            ('PD', month),
            ('PY', year),
            ('VL', volume),
            ('IS', issue),
            ('PN', partNumber),
            ('SU', supplement),
            ('SI', specialIssue),
            ('MA', meetingAbstract),
            ('BP', beginningPage),
            ('EP', endingPage),
            ('AR', articleNumber),
            ('PG', pageCount),
            ('WC', subjects),
            ('DI', DOI),
            ('D2', bookDOI),
            ('GA', documentDeliveryNumber),
            ('UT', wosString),
            ('PM', pubMedID),
            ])



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016
import unittest
import metaknowledge

class TestWOS(unittest.TestCase):
    def setUp(self):
        self.R = metaknowledge.WOSRecord(simplePaperString)
        self.Rbad = metaknowledge.WOSRecord(simplePaperString[:-3])

    def test_creation(self):
        R = metaknowledge.WOSRecord(self.R._fieldDict)
        self.assertEqual(R, self.R)
        with open("metaknowledge/tests/testFile.isi") as f:
            f.readline()
            f.readline()
            R = metaknowledge.WOSRecord(f)
            self.assertEqual(R.id, 'WOS:A1979GV55600001')
        with self.assertRaises(TypeError):
            R = metaknowledge.WOSRecord(123456789)

    def test_badwrite(self):
        with self.assertRaises(metaknowledge.BadWOSRecord):
            self.Rbad.writeRecord('not a file object.txt')

    def test_dupDetection(self):
        s = simplePaperString[:-3] + "DE Example; testing\nPD APR\nER\n"
        R = metaknowledge.WOSRecord(s)
        self.assertTrue(R.bad)

    def test_WOSNum(self):
        self.assertEqual(self.R.UT, 'WOS:123317623000007')
        self.assertEqual(self.R.wosString, 'WOS:123317623000007')


simplePaperString = """PT J
AU John, D
AF John, Doe
TI Example Paper
SO TOPICS IN COGNITIVE SCIENCE
LA English
DT Article
DE Example; testing
ID REAL; TIME
AB This is a test.
C1 UW, Ontario, Canada.
RP John, D (reprint author), UW, Ontario, Canada.
CR John D. 1999, TOPICS IN COGNITIVE SCIENCE
J9 EXAMPLE
JI examaple
PD APR
PY 2015
VL 1
BP 1
EP 2
DI 10.1111
UT WOS:123317623000007
ER
"""



import collections

from .baseGrant import Grant
from ..mkExceptions import BadGrant

class MedlineGrant(Grant):
    def __init__(self, grantString):

        grantDict = collections.OrderedDict()

        bad = False
        error = None

        split = grantString.split('/')
        try:
            grantDict['country'] = split.pop()
            grantDict['agency'] = split.pop()
        except IndexError:
            bad = True
            grantDict['country'] = grantString
            error = BadGrant("The grant string '{}' does not contain enough slashes (/) to be a medline grant.".format(grantString))
        else:
            if len(split) == 1:
                code = split.pop()
                if len(code) == 2:
                    grantDict['code'] = code
                else:
                    grantDict['number'] = code
            elif len(split) == 2:
                code = split.pop()
                if len(code) == 2:
                    grantDict['code'] = code
                    grantDict['number'] = split.pop()
                else:
                    grantDict['number'] = "{}/{}".format(split.pop(), code)
            else:
                grantDict['number'] = '/'.join(split)
        if 'number' in  grantDict:
            idValue = "{}/{}-{}".format(grantDict.get('number', ''), grantDict.get('code', ''), grantDict.get('country', ''))
        else:
            idValue  = "{}-{}-{}".format(grantDict.get('code', ''), grantDict.get('agency', ''), grantDict.get('country', ''))
        Grant.__init__(self, grantString, grantDict, idValue, bad, error)



import shutil
import collections
import datetime

from .mkExceptions import mkException

glimpseTags = collections.OrderedDict([
    ('Top Authors','authorsFull'),
    ('Top Journals','journal'),
    ('Top Cited','citations'),
    ])

descriptionString1 = 'Columns are ranked by num. of occurrences'
descriptionString2 = 'and are independent of one another'

descriptionStringFull = descriptionString1 + ' ' + descriptionString2

def _glimpse(RC, *tags, compact = False):
    tColumns, tRows = tuple(shutil.get_terminal_size())
    if len(tags) < 1:
        targetTags = glimpseTags
    else:
        targetTags = {t: t for t in tags}
    #If it can't fit just go with the usual settings
    if tColumns < 55:
        tColumns = 80
    if tRows < 6:
        tRows = 24
    glimpseVals = collections.OrderedDict()
    if len(descriptionStringFull) > tColumns:
        maxRows = tRows - 7
    else:
        maxRows = tRows - 6
    for name, tag in targetTags.items():
        glimpseVals[name] = RC.rankedSeries(tag, giveCounts = False, giveRanks = True, pandasMode = False)
    return makeHeader(RC, tColumns, targetTags, compact) + makeTable(glimpseVals, maxRows, tColumns, compact)

def makeHeader(RC, width, glimpseVals, compact):
    now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    firstLine = "{} glimpse made at: {}".format(type(RC).__name__, now)
    secondLine = "{} Records from {}".format(len(RC), RC.name[:30])
    if compact:
        if len(descriptionStringFull) > width - 2:
            thirdLine = '|{1:+<{0}}|\n|{2:+<{0}}|\n'.format(width - 2, descriptionString1, descriptionString2)
        else:
            thirdLine = '|{1:+<{0}}|\n'.format(width - 2, descriptionStringFull)
        return '+{2:+<{0}}\n|{3:+<{1}}|\n{4}'.format(width - 1, width - 2, firstLine, secondLine, thirdLine)
    else:
        return '{}\n{}\n'.format(firstLine, secondLine)

def makeTable(values, height, width, compact):
    retLines = []
    if compact:
        lines = [[] for i in range(height + 1)]
        firstRowString = "|{}" + "+{}" * (len(values) - 1) + '|'
        rowString = "|{}" * len(values) + '|'
        cWidth = (width // len(values)) - 1
        cRemainder = width % len(values) - 1
        for title, rows in values.items():
            if cRemainder > 0:
                heading = "{1:-^{0}}".format(cWidth + 1, title)
                cRemainder -= 1
            elif cRemainder < 0:
                heading = "{1:-^{0}}".format(cWidth - 1, title)
                cRemainder += 1
            else:
                heading = "{1:-^{0}}".format(cWidth, title)
            hWidth = len(heading)
            lines[0].append(heading)
            if len(rows) < height:
                for i in range(height - len(rows)):
                    rows.append(('NA', -1))
            for index, entry in enumerate((prepEntry(hWidth, *s) for s in rows[:height]), start = 1):
                lines[index].append(entry)
        retLines.append(firstRowString.format(*tuple(lines[0])))
        for line in lines[1:]:
            retLines.append(rowString.format(*tuple(line)))
    else:
        for title, rows in values.items():
            retLines.append('')
            retLines.append(title)
            retLines += ['{} {}'.format(c, str(s)[:width - len(str(c)) - 1]) for s, c in rows[:height // 2]]
    return '\n'.join(retLines)

def prepEntry(maxLength, valString, rank):
    valString = str(valString)
    if len(valString) <= maxLength - 2:
        valString = valString.rjust(maxLength - 2, ' ')
    else:
        valString = "{}.".format(valString[:maxLength - 3])
    return "{:<2.0f}{}".format(rank, valString)



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016
"""These are the functions used to process scopus csv files at the backend. They are meant for use internal use by metaknowledge.
"""
from .recordScopus import ScopusRecord, scopusRecordParser, scopusHeader
from .scopusHandlers import isScopusFile, scopusParser

from .tagProcessing.tagFunctions import scopusTagToFunction
from .tagProcessing.specialFunctions import scopusSpecialTagToFunc



from ...grants.scopusGrant import ScopusGrant
from ...citation import Citation

def commaSpaceSeperated(val):
    return val.split(', ')

def semicolonSpaceSeperated(val):
    return val.split('; ')

def semicolonSeperated(val):
    return val.split(';')

def stringValue(val):
    return val

def integralValue(val):
    return int(val)

def grantValue(val):
    return [ScopusGrant(s) for s in val.split('; ')]

def citeValue(val):
    return [Citation(s, scopusMode = True) for s in val.split('; ')]

scopusTagToFunction = {
    'Authors' : commaSpaceSeperated,
    'Title' : stringValue,
    'Year' : integralValue,
    'Source title' : stringValue,
    'Volume' : stringValue,
    'Issue' : stringValue,
    'Art. No.' : stringValue,
    'Page start' : stringValue,
    'Page end' : stringValue,
    'Page count' : integralValue,
    'Cited by' : integralValue,
    'DOI' : stringValue,
    'Link' : stringValue,
    'Affiliations' : stringValue,
    'Authors with affiliations' : semicolonSpaceSeperated,
    'Abstract' : stringValue,
    'Author Keywords' : semicolonSpaceSeperated,
    'Index Keywords' : semicolonSpaceSeperated,
    'Molecular Sequence Numbers' : stringValue,
    'Chemicals/CAS' : stringValue,
    'Tradenames' : semicolonSpaceSeperated,
    'Manufacturers' : semicolonSpaceSeperated,
    'Funding Details' : grantValue,
    'References' : citeValue,
    'Correspondence Address' : semicolonSpaceSeperated,
    'Editors' : stringValue,
    'Sponsors' : semicolonSeperated,
    'Publisher' : stringValue,
    'Conference name' : stringValue,
    'Conference date' : stringValue,
    'Conference location' : stringValue,
    'Conference code' : integralValue,
    'ISSN' : stringValue,
    'ISBN' : semicolonSpaceSeperated,
    'CODEN' : stringValue,
    'PubMed ID' : stringValue,
    'Language of Original Document' : semicolonSpaceSeperated,
    'Abbreviated Source Title' : stringValue,
    'Document Type' : stringValue,
    'Source' : stringValue,
    'EID' : stringValue,

}



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016

import collections
import csv
import re

from .tagProcessing.tagFunctions import scopusTagToFunction
from .tagProcessing.specialFunctions import scopusSpecialTagToFunc

from ..mkRecord import ExtendedRecord
from ..mkExceptions import RCTypeError, BadScopusFile, BadScopusRecord

scopusHeader = [
    'Authors',
    'Title',
    'Year',
    'Source title',
    'Volume',
    'Issue',
    'Art. No.',
    'Page start',
    'Page end',
    'Page count',
    'Cited by',
    'DOI',
    'Link',
    'Affiliations',
    'Authors with affiliations',
    'Abstract',
    'Author Keywords',
    'Index Keywords',
    'Molecular Sequence Numbers',
    'Chemicals/CAS',
    'Tradenames',
    'Manufacturers',
    'Funding Details',
    'Funding Text',
    'References',
    'Correspondence Address',
    'Editors',
    'Sponsors',
    'Publisher',
    'Conference name',
    'Conference date',
    'Conference location',
    'Conference code',
    'ISSN',
    'ISBN',
    'CODEN',
    'PubMed ID',
    'Language of Original Document',
    'Abbreviated Source Title',
    'Document Type',
    'Source',
    'EID'
 ]

class ScopusRecord(ExtendedRecord):
    """Class for full Scopus entries.

    This class is an [ExtendedRecord](./ExtendedRecord.html#metaknowledge.ExtendedRecord) capable of generating its own id number. You should not create them directly, but instead use [scopusParser()](../modules/scopus.html#metaknowledge.scopus.scopusHandlers.scopusParser) on a scopus **CSV** file.
    """
    def __init__(self, inRecord, sFile = "", sLine = 0, header = None):
        bad = False
        error = None
        fieldDict = None
        try:
            if isinstance(inRecord, dict) or isinstance(inRecord, collections.OrderedDict):
                fieldDict = collections.OrderedDict(inRecord)
            elif isinstance(inRecord, str):
                fieldDict = scopusRecordParser(inRecord, header = header)
            else:
                raise RCTypeError("Unsupported input type '{}', ScopusRecords cannot be created from '{}'".format(inRecord, type(inRecord)))
        except (BadScopusRecord, IndexError) as b:
            self.bad = True
            self.error = b
            fieldDict = collections.OrderedDict()
        if fieldDict is not None:
            if 'EID' in fieldDict:
                self._scopusNum = "EID:{}".format(fieldDict['EID'])
            else:
                self._scopusNum = None
                bad = True
                error = BadScopusRecord("Missing EID")
        ExtendedRecord.__init__(self, fieldDict, self._scopusNum, bad, error, sFile = sFile, sLine = sLine)

    def encoding(self):
        return 'utf-8'

    @staticmethod
    def getAltName(tag):
        return None

    @staticmethod
    def tagProcessingFunc(tag):
        return scopusTagToFunction[tag]

    def specialFuncs(self, key):
        return scopusSpecialTagToFunc[key](self)

    def writeRecord(self, f):
        if self.bad:
            raise BadScopusRecord("This record cannot be converted to a file as the input was malformed.\nThe original line number (if any) is: {} and the original file is: '{}'".format(self._sourceLine, self._sourceFile))
        else:
            f.write(','.join(('"{}"'.format(self._fieldDict.get(k, '')) for k in scopusHeader)))

    def createCitation(self, multiCite = False):
        """Overwriting the general [citation creator](./ExtendedRecord.html#metaknowledge.ExtendedRecord.createCitation) to deal with scopus weirdness.

        Creates a citation string, using the same format as other WOS citations, for the [Record](./Record.html#metaknowledge.Record) by reading the relevant special tags (`'year'`, `'J9'`, `'volume'`, `'beginningPage'`, `'DOI'`) and using it to create a [Citation](./Citation.html#metaknowledge.citation.Citation) object.

        # Parameters

        _multiCite_ : `optional [bool]`

        > Default `False`, if `True` a tuple of Citations is returned with each having a different one of the records authors as the author

        # Returns

        `Citation`

        > A [Citation](./Citation.html#metaknowledge.citation.Citation) object containing a citation for the Record.
        """
        #Need to put the import here to avoid circular import issues
        from ..citation import Citation
        valsStr = ''
        if multiCite:
            auths = []
            for auth in self.get("authorsShort", []):
                auths.append(auth.replace(',', ''))
        else:
            if self.get("authorsShort", False):
                valsStr += self['authorsShort'][0].replace(',', '') + ', '
        if self.get("title", False):
            valsStr += self.get('title').replace('(', '').replace(')', '') + ' '
        if self.get("year", False):
            valsStr += "({}) ".format(self.get('year'))
        if self.get("journal", False):
            valsStr += self.get('journal') + ', '
        if self.get("volume", False):
            valsStr += str(self.get('volume')) + ', '
        if self.get("beginningPage", False):
            valsStr += 'PP. ' + str(self.get('beginningPage'))
        if multiCite and len(auths) > 0:
            ret = (tuple((Citation(a + valsStr, scopusMode = True) for a in auths)))
        elif multiCite:
            ret = Citation(valsStr, scopusMode = True),
        else:
            ret = Citation(valsStr, scopusMode = True)
        if multiCite:
            rL = []
            for c in ret:
                if c.bad:
                    c.year = self.get('year', 0)
                    c.name = self.get('title', '').upper()
                    c.journal = self.get("journal", '').upper()
                rL.append(c)
            return tuple(rL)
        else:
            if ret.bad:
                ret.year = self.get('year', 0)
                ret.name = self.get('title', '').upper()
                ret.journal = self.get("journal", '').upper()
            return ret

firstQuotingRegex = re.compile(r'("")*"([^"]|"$)')
innerQuotingRegex = re.compile(r'("")*"([^"|$])')

def scopusRecordParser(record, header = None):
    """The parser [ScopusRecords](../classes/ScopusRecord.html#metaknowledge.scopus.ScopusRecord) use. This takes a line from [scopusParser()](#metaknowledge.scopus.scopusHandlers.scopusParser) and parses it as a part of the creation of a `ScopusRecord`.

    **Note** this is for csv files downloaded from scopus _not_ the text records as those are less complete. Also, Scopus uses double quotes (`"`) to quote strings, such as abstracts, in the csv so double quotes in the string must be escaped. For reasons not fully understandable by mortals they choose to use two double quotes in a row (`""`) to represent an escaped double quote. This parser does not unescape these quotes, but it does correctly handle their interacts with the outer double quotes.

    # Parameters

    _record_ : `str`

    > string ending with a newline containing the record's entry

    # Returns

    `dict`

    > A dictionary of the key-vaue pairs in the entry
    """
    if header is None:
        header = scopusHeader
    splitRecord = record[:-1].split(',')
    tagDict = {}
    quoted = False
    for key in reversed(header):
        currentVal = splitRecord.pop()
        if currentVal == '':
            pass
        elif currentVal[-1] == '"':
            if re.match(firstQuotingRegex, currentVal) is None:
                valString = ',' + currentVal[:-1]
                currentVal = splitRecord.pop()
                #double quotes (") are escaped by proceeding them with another double quote
                #So an entry containing:

                #',"stuff,""quoted"",more stuff,""more quoted""",'

                #would be a single string belonging to 1 column that looks like:

                #'stuff,"quoted",more stuff,"more quoted"'

                #We are not going to unescape the quotation marks but we do have to deal with them
                while re.match(innerQuotingRegex, currentVal) is None:
                    valString = ',' + currentVal + valString
                    currentVal = splitRecord.pop()
                valString = currentVal[1:] + valString
            else:
                try:
                    valString = currentVal[1:-1]
                except ValueError:
                    valString = currentVal[1:-1]
            tagDict[key] = valString
        else:
            tagDict[key] = currentVal
    return tagDict



import collections

from .baseGrant import Grant
from ..mkExceptions import BadGrant

class ScopusGrant(Grant):
    def __init__(self, grantString):

        grantDict = collections.OrderedDict()

        bad = False
        error = None

        split = grantString.split(', ')
        try:
            grantDict['agency'] = split.pop()
        except IndexError:
            bad = True
            grantDict['agency'] = grantString
            error = BadGrant("The grant string '{}' does not contain enough comma-spaces (', ') to be a scopus grant.".format(grantString))
        else:
            try:
                grantDict['agencyCode'] = split.pop()
            except IndexError:
                pass
            else:
                try:
                    grantDict['code'] = split.pop()
                except IndexError:
                    pass
        idValue  = "{}-{}-{}".format(grantDict.get('code', ''), grantDict.get('agencyCode', ''), grantDict.get('agency', ''))
        Grant.__init__(self, grantString, grantDict, idValue, bad, error)



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016
"""These are the functions used to process medline (pubmed) files at the backend. They are meant for use internal use by metaknowledge.
"""
from .recordProQuest import ProQuestRecord, proQuestRecordParser
from .proQuestHandlers import isProQuestFile, proQuestParser
from .tagProcessing.specialFunctions import proQuestSpecialTagToFunc
from .tagProcessing.tagFunctions import proQuestTagToFunc



import sys
import os
import os.path
import argparse
import builtins
import code
import unittest
import unittest.mock

import metaknowledge
import metaknowledge.bin.metaknowledgeCLI

class MockInput(unittest.mock.MagicMock):

    def __init__(self, *args, **kwargs):
        unittest.mock.MagicMock.__init__(self,*args, **kwargs)
        self.calledVals = []

    def __call__(self, *args, **kwargs):
        val = self.calledVals.pop(0)
        if not isinstance(val, str):
            raise val
        else:
            return val

class TestCLI(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.sysArgs = sys.argv
        cls.RCmain = metaknowledge.RecordCollection("metaknowledge/tests/testFile.isi")
        cls.Gmain = cls.RCmain.networkCoAuthor()

    def setUp(self):
        self.RC = self.RCmain.copy()
        self.G = self.Gmain.copy()
        sys.argv = ['/usr/local/bin/metaknowledge']

    #Trying to make nose happy
    def tearDown(self):
        sys.argv = self.sysArgs

    def test_quit(self):
        sys.argv = ['/usr/local/bin/metaknowledge', '--debug', '--quiet', 'metaknowledge/tests']
        with unittest.mock.patch('builtins.print'):
            with unittest.mock.patch('builtins.input', return_value='q'):
                with self.assertRaises(SystemExit):
                    metaknowledge.bin.mkCLI()

    def test_startInteractive(self):
        sys.argv = ['/usr/local/bin/metaknowledge', '--debug', '--quiet', 'metaknowledge/tests']
        with unittest.mock.patch('builtins.print'):
            with unittest.mock.patch('builtins.input', new_callable = MockInput) as m:
                m.calledVals = [KeyboardInterrupt]
                self.assertEqual(metaknowledge.bin.mkCLI(), 0)
                self.assertEqual(len(m.calledVals), 0)

    def test_keyboardInterupt(self):
        sys.argv = ['/usr/local/bin/metaknowledge', '--debug', '--quiet', 'metaknowledge/tests']
        with unittest.mock.patch('builtins.print'):

            class mockExit(unittest.mock.MagicMock):
                def __call__(self, *args, **kwargs):
                    #Just want to raise an error, --debug mode means it will not be caught
                    raise RuntimeError

            with unittest.mock.patch('code.interact', new_callable = mockExit):
                with unittest.mock.patch('builtins.input', new_callable = MockInput) as m:
                    m.calledVals = ['i']
                    with self.assertRaises(RuntimeError):
                        metaknowledge.bin.mkCLI()
                    self.assertEqual(len(m.calledVals), 0)

    def test_basicRun(self):
        sys.argv = ['/usr/local/bin/metaknowledge', '--debug', '--quiet', 'metaknowledge/tests']
        with unittest.mock.patch('builtins.print'):
            with unittest.mock.patch('builtins.input', new_callable = MockInput) as m:
                fileName = 'CLITempFile.csv'
                f = open(fileName, 'w')
                f.write("tempFile for testing metaknowledge should have been deleted. A test went wrong.")
                f.close()
                m.calledVals = ['1', '6', '0', '2', fileName[:-4], 'p','y']
                self.assertEqual(metaknowledge.bin.mkCLI(), 1)
                self.assertEqual(os.path.getsize(fileName), 4110112)
                self.assertEqual(len(m.calledVals), 0)
                os.remove(fileName)

    def test_spamming(self):
        with unittest.mock.patch('builtins.print'):
            with unittest.mock.patch('builtins.input', return_value = '0') as m:
                self.assertIsInstance(metaknowledge.bin.mkCLI(), Exception)

    def test_Thresholds(self):
        with unittest.mock.patch('builtins.print'):
            with unittest.mock.patch('builtins.input', new_callable = MockInput) as m:
                m.calledVals = ['', '0']
                G = metaknowledge.bin.metaknowledgeCLI.getThresholds(None, self.G)
                self.assertEqual(G, self.G)
                self.assertEqual(len(m.calledVals), 0)

                m.calledVals = ['1', '2', '3', '1', '4', '5', '5', '0', '6', 'ten', '2', '0']
                G = metaknowledge.bin.metaknowledgeCLI.getThresholds(None, self.G)
                self.assertEqual(metaknowledge.graphStats(G, sentenceString = True), 'The graph has 2 nodes, 1 edges, 0 isolates, 0 self loops, a density of 1 and a transitivity of 0')
                self.assertEqual(len(m.calledVals), 0)

    def test_graphs(self):
        with unittest.mock.patch('builtins.print'):
            with unittest.mock.patch('builtins.input', new_callable = MockInput) as m:
                m.calledVals = ['1', 'NOT A TAG', 'AF']
                G = metaknowledge.bin.metaknowledgeCLI.getNetwork(None, self.RC)
                self.assertEqual(metaknowledge.graphStats(G, sentenceString = True), 'The graph has 45 nodes, 46 edges, 9 isolates, 0 self loops, a density of 0.0464646 and a transitivity of 0.822581')
                self.assertEqual(len(m.calledVals), 0)

                m.calledVals = ['2', 'AF', 'UT']
                G = metaknowledge.bin.metaknowledgeCLI.getNetwork(None, self.RC)
                self.assertEqual(metaknowledge.graphStats(G, sentenceString = True), 'The graph has 77 nodes, 66 edges, 0 isolates, 0 self loops, a density of 0.0225564 and a transitivity of 0')
                self.assertEqual(len(m.calledVals), 0)

                m.calledVals = ['3'] + list(self.RC.tags()) + ['']
                G = metaknowledge.bin.metaknowledgeCLI.getNetwork(None, self.RC)
                self.assertEqual(metaknowledge.graphStats(G, sentenceString = True), 'The graph has 1186 nodes, 38564 edges, 0 isolates, 56 self loops, a density of 0.0549192 and a transitivity of 0.295384')
                self.assertEqual(len(m.calledVals), 0)

                m.calledVals = ['4']
                G = metaknowledge.bin.metaknowledgeCLI.getNetwork(None, self.RC)
                self.assertEqual(metaknowledge.graphStats(G, sentenceString = True), 'The graph has 511 nodes, 817 edges, 0 isolates, 0 self loops, a density of 0.00313495 and a transitivity of 0.00600437')
                self.assertEqual(len(m.calledVals), 0)

                m.calledVals = ['5']
                G = metaknowledge.bin.metaknowledgeCLI.getNetwork(None, self.RC)
                self.assertEqual(metaknowledge.graphStats(G, sentenceString = True), 'The graph has 493 nodes, 13000 edges, 0 isolates, 22 self loops, a density of 0.107282 and a transitivity of 0.611431')
                self.assertEqual(len(m.calledVals), 0)

                m.calledVals = ['6']
                G = metaknowledge.bin.metaknowledgeCLI.getNetwork(None, self.RC)
                self.assertEqual(metaknowledge.graphStats(G, sentenceString = True), 'The graph has 45 nodes, 46 edges, 9 isolates, 0 self loops, a density of 0.0464646 and a transitivity of 0.822581')
                self.assertEqual(len(m.calledVals), 0)

    def test_create(self):
        fileName = 'tempTestFile'
        named = argparse.Namespace()
        named.name = fileName
        unnamed = argparse.Namespace()
        unnamed.name = None
        with unittest.mock.patch('builtins.print'):
            with unittest.mock.patch('builtins.input', new_callable = MockInput) as m:
                m.calledVals = ['1']
                self.assertTrue(metaknowledge.bin.metaknowledgeCLI.getWhatToDo(named, self.RC))
                self.assertEqual(len(m.calledVals), 0)

                m.calledVals = ['2']
                self.RC.dropBadEntries()
                self.assertFalse(metaknowledge.bin.metaknowledgeCLI.getWhatToDo(named, self.RC))
                self.assertEqual(os.path.getsize(fileName+ '.txt'), 88160)
                self.assertEqual(len(m.calledVals), 0)

                m.calledVals = ['3', fileName, 'n', fileName, 'g', 'y']
                self.RC.dropBadEntries()
                self.assertTrue(metaknowledge.bin.metaknowledgeCLI.getWhatToDo(unnamed, self.RC))
                self.assertEqual(os.path.getsize(fileName+ '.txt'), 88160)
                os.remove(fileName + '.txt')
                self.assertEqual(len(m.calledVals), 0)

                m.calledVals = ['4']
                self.RC.dropBadEntries()
                self.assertFalse(metaknowledge.bin.metaknowledgeCLI.getWhatToDo(named, self.RC))
                self.assertEqual(os.path.getsize(fileName+ '.csv'), 86330)
                os.remove(fileName + '.csv')
                self.assertEqual(len(m.calledVals), 0)

                m.calledVals = ['6', 'y']
                self.RC.dropBadEntries()
                self.assertFalse(metaknowledge.bin.metaknowledgeCLI.getWhatToDo(named, self.RC))
                self.assertEqual(os.path.getsize(fileName+ '.csv'), 20123)
                os.remove(fileName + '.csv')
                self.assertEqual(len(m.calledVals), 0)

                with self.assertRaises(KeyboardInterrupt):
                    #Don't want to mess with these too much
                    m.calledVals = ['7', '', KeyboardInterrupt]
                    self.assertFalse(metaknowledge.bin.metaknowledgeCLI.getWhatToDo(named, self.RC))
                self.assertEqual(len(m.calledVals), 0)
                m.calledVals = ['7', '']
                self.assertFalse(metaknowledge.bin.metaknowledgeCLI.getWhatToDo(named, metaknowledge.RecordCollection()))
                self.assertEqual(len(m.calledVals), 0)

    def test_output(self):
        fileName = 'tempTestFile'
        named = argparse.Namespace()
        named.name = fileName
        unnamed = argparse.Namespace()
        unnamed.name = None
        with unittest.mock.patch('builtins.print'):
            with unittest.mock.patch('builtins.input', new_callable = MockInput) as m:
                f = open(fileName + '_edgeList.csv', 'w')
                f.write("tempFile for testing metaknowledge should have been deleted. A test went wrong.")
                f.close()


                m.calledVals = ['1']
                metaknowledge.bin.metaknowledgeCLI.outputNetwork(named, self.G)
                #filesize is slightly non deteministic
                self.assertAlmostEqual(os.path.getsize(fileName + '_edgeList.csv'), 1403, delta = 3)
                self.assertAlmostEqual(os.path.getsize(fileName + '_nodeAttributes.csv'), 779, delta = 4)
                os.remove(fileName + '_nodeAttributes.csv')
                os.remove(fileName + '_edgeList.csv')
                self.assertEqual(len(m.calledVals), 0)

                m.calledVals = ['4']
                metaknowledge.bin.metaknowledgeCLI.outputNetwork(named, self.G)
                #filesize is slightly non deteministic
                self.assertAlmostEqual(os.path.getsize(fileName + '.graphml'), 6817, delta = 4)
                os.remove(fileName + '.graphml')
                self.assertEqual(len(m.calledVals), 0)



from .nsercGrant import NSERCGrant, isNSERCfile, parserNSERCfile
from .medlineGrant import MedlineGrant
from .baseGrant import Grant, FallbackGrant, isFallbackGrantFile, parserFallbackGrantFile
from .cihrGrant import CIHRGrant, isCIHRfile, parserCIHRfile
from .nsfGrant import NSFGrant, isNSFfile, parserNSFfile


"""#Creating new grants

mk is intended to be expanded as different researchers will require processing of files not in it currently. To add a new grant you need to write 2 simple functions and class. To see a basic example look at the `baseGrant.py` file as it contains the fallback grant processors and you should be able to reuse much of that code.

The way GrantCollections are created is when they are given a file or directory of files they check each file with the `detector` functions in the `grantProcessors` found in `fileHandlers.py` and if one returns `True` they use the `processor` function and added the `type` string to their `collectedTypes` set. The `processor` must return a tuple the first element being a set of all the Grants the second `None` or an `Exception` object. `processor` should not raise an exception, if there is an issue the GrantCollection should be given even a partial set of grants, GrantCollections have an errors attribute that contains all errors they encountered during the parsing.


The first function is to determine if a given file path is to a collection of grants of the needed type. Determining if a file is of the needed type is usually done by reading the first few lines and checking that they match a known header template. For example CIHR files start with the string `"Search Criteria,"` so the function `isCIHRfile()` checks that the first lines start that way.

One thing to watch out for is the encoding most grants are CSVs encoded with ISO-8859 which is what many windows programs, most notably excel, expect. Python will use that encoding (called `'latin-1'`) on Microsoft systems but on Mac OS and Linux will often use `'utf-8'` so you should always give the encoding as mk is intended for all 3 operating systems.

The next function is the parser, this is the function that is called on the file to create the Grants. It is given a file path that has been confirmed to be a correctly formatted grant by the detector.

The function must return a tuple the first entry being a set of all the Grants and the second an `Exception` if an error occurred or `None` if not. If an error occurs the function should attempt to return as many grants as possible, including the one that had the error (with its error handlers correctly indicated). The GrantCollectio will record the error and allow for the user/script to decide what actions to take.Note, often not doing anything is appropriate as errors have been found to most often occur at the end of the file so no data is actually lost.

The`Grants` in the set returned by the processor, should be a new class the inherits from `Grant` even if no new attributes are defined.

Once the `detector` and `processor` functions have been created and tested, they can be added to the list of grants found in `fileHandlers.py` called `grantProcessors`. Each file is checked in order so do not add them to the end as the last entry will tell the `GrantCollection` to stop and that the file does not match, so placing anything after it will not work.
"""



from ...grants.medlineGrant import MedlineGrant

def IR(val):
    """Investigator"""
    return val

def FIR(val):
    """InvestigatorFull"""
    return val

def PROF(val):
    """PartialRetractionOf"""
    return val

def PUBM(val):
    """PublishingModel"""
    return val

def CN(val):
    """CorporateAuthor"""
    return val

def MHDA(val):
    """MeSHDate"""
    return val

def LID(val):
    """LocationIdentifier"""
    return val

def EDAT(val):
    """EntrezDate"""
    return val

def OCI(val):
    """OtherCopyright"""
    return val

def SB(val):
    """Subset"""
    return val

def DA(val):
    """DateCreated"""
    return val

def PMCR(val):
    """PubMedCentralRelease"""
    return val

def PG(val):
    """Pagination
    all pagination seen so far seems to be only on one line"""
    return val[0]

def GS(val):
    """GeneSymbol"""
    return val

def VI(val):
    """Volume
    The volumes as a string as volume is single line
    """
    return val[0]

def UOF(val):
    """UpdateOf"""
    return val

def OWN(val):
    """Owner"""
    return val

def ORI(val):
    """OriginalReportIn"""
    return val

def MID(val):
    """ManuscriptIdentifier"""
    return val

def PMID(val):
    """PubMedUniqueIdentifier"""
    return "PMID:{}".format(val[0])

def PMC(val):
    """PubMedCentralIdentifier"""
    return val

def RIN(val):
    """RetractionIn"""
    return val

def RPF(val):
    """RepublishedFrom"""
    return val

def CIN(val):
    """CommentIn"""
    return val

def FPS(val):
    """FullPersonalNameSubject"""
    return val

def TT(val):
    """TransliteratedTitle"""
    return val

def PHST(val):
    """PublicationHistoryStatus"""
    return val

def EFR(val):
    """ErratumFor"""
    return val

def PST(val):
    """PublicationStatus"""
    return val

def SPIN(val):
    """SummaryForPatients"""
    return val

def AU(val):
    """Author"""
    return val

def FED(val):
    """Editor"""
    return val

def NM(val):
    """SubstanceName"""
    return val

def SO(val):
    """Source"""
    return val

def IP(val):
    """Issue"""
    return val

def OABL(val):
    """OtherAbstract"""
    return val

def CRDT(val):
    """CreateDate"""
    return val

def DDIN(val):
    """DatasetIn"""
    return val

def MH(val):
    """MeSHTerms"""
    return val

def DP(val):
    """DatePublication"""
    return val[0]

def GN(val):
    """GeneralNote"""
    return val

def CRF(val):
    """CorrectedRepublishedFrom"""
    return val

def TI(val):
    """Title
    only one per record"""
    return val[0]

def CRI(val):
    """CorrectedRepublishedIn"""
    return val

def OT(val):
    """OtherTerm
    Nothing needs to be done"""
    return val

def ROF(val):
    """RetractionOf"""
    return val

def OTO(val):
    """OtherTermOwner
    one line field"""
    return val[0]

def OID(val):
    """OtherID"""
    return val

def PT(val):
    """PublicationType"""
    return val

def RPI(val):
    """RepublishedIn"""
    return val

def AB(val):
    """Abstract
    basically a one liner after parsing"""
    return val[0]

def EN(val):
    """Edition"""
    return val

def AD(val):
    """Affiliation
    Undoing what the parser does then splitting at the semicolons and dropping newlines extra fitlering is required beacuse some AD's end with a semicolon"""
    retDict = {}
    for v in val:
        split = v.split(' : ')
        retDict[split[0]] = [s for s in' : '.join(split[1:]).replace('\n', '').split(';') if s != '']
    return retDict

def LA(val):
    """Language"""
    return val

def TA(val):
    """JournalTitleAbbreviation
    One line only"""
    return val[0]

def JT(val):
    """JournalTitle
    One line only"""
    return val[0]

def IRAD(val):
    """InvestigatorAffiliation"""
    return val

def PS(val):
    """PersonalNameSubject"""
    return val

def IS(val):
    """ISSN"""
    return val

def PL(val):
    """PlacePublication"""
    return val

def CTI(val):
    """CollectionTitle"""
    return val

def FAU(val):
    """FullAuthor"""
    return val

def VTI(val):
    """VolumeTitle"""
    return val

def DCOM(val):
    """DateCompleted"""
    return val

def BTI(val):
    """BookTitle"""
    return val

def CI(val):
    """CopyrightInformation"""
    return val

def STAT(val):
    """Status"""
    return val

def DRIN(val):
    """DatasetUseReportedIn"""
    return val

def RF(val):
    """NumberReferences"""
    return val

def UIN(val):
    """UpdateIn"""
    return val

def LR(val):
    """DateLastRevised"""
    return val

def SFM(val):
    """SpaceFlightMission"""
    return val

def EIN(val):
    """ErratumIn"""
    return val

def AID(val):
    """ArticleIdentifier
    The given values do not require any work"""
    return val

def PRIN(val):
    """PartialRetractionIn"""
    return val

def DEP(val):
    """DateElectronicPublication"""
    return val

def AUID(val):
    """AuthorIdentifier
    one line only just need to undo the parser's effects"""
    retDict = {}
    for v in val:
        split = v.split(' : ')
        retDict[split[0]] = ' : '.join(split[1:])
    return retDict

def SI(val):
    """SecondarySourceID"""
    return val

def ISBN(val):
    """ISBN"""
    return val

def RN(val):
    """RegistryNumber"""
    return val

def JID(val):
    """NLMID"""
    return val

def GR(val):
    """GrantNumber"""
    return [MedlineGrant(v) for v in val]

medlineTagToFunc = {
    "IR" : IR,
    "FIR" : FIR,
    "PROF" : PROF,
    "PUBM" : PUBM,
    "CN" : CN,
    "MHDA" : MHDA,
    "LID" : LID,
    "EDAT" : EDAT,
    "OCI" : OCI,
    "SB" : SB,
    "DA" : DA,
    "PMCR" : PMCR,
    "PG" : PG,
    "GS" : GS,
    "VI" : VI,
    "UOF" : UOF,
    "OWN" : OWN,
    "ORI" : ORI,
    "MID" : MID,
    "PMID" : PMID,
    "PMC" : PMC,
    "RIN" : RIN,
    "RPF" : RPF,
    "CIN" : CIN,
    "FPS" : FPS,
    "TT" : TT,
    "PHST" : PHST,
    "EFR" : EFR,
    "PST" : PST,
    "SPIN" : SPIN,
    "AU" : AU,
    "FED" : FED,
    "NM" : NM,
    "SO" : SO,
    "IP" : IP,
    "OABL" : OABL,
    "CRDT" : CRDT,
    "DDIN" : DDIN,
    "MH" : MH,
    "DP" : DP,
    "GN" : GN,
    "CRF" : CRF,
    "TI" : TI,
    "CRI" : CRI,
    "OT" : OT,
    "ROF" : ROF,
    "OTO" : OTO,
    "OID" : OID,
    "PT" : PT,
    "RPI" : RPI,
    "AB" : AB,
    "EN" : EN,
    "AD" : AD,
    "LA" : LA,
    "TA" : TA,
    "JT" : JT,
    "IRAD" : IRAD,
    "PS" : PS,
    "IS" : IS,
    "PL" : PL,
    "CTI" : CTI,
    "FAU" : FAU,
    "VTI" : VTI,
    "DCOM" : DCOM,
    "BTI" : BTI,
    "CI" : CI,
    "STAT" : STAT,
    "DRIN" : DRIN,
    "RF" : RF,
    "UIN" : UIN,
    "LR" : LR,
    "SFM" : SFM,
    "EIN" : EIN,
    "AID" : AID,
    "PRIN" : PRIN,
    "DEP" : DEP,
    "AUID" : AUID,
    "SI" : SI,
    "ISBN" : ISBN,
    "RN" : RN,
    "JID" : JID,
    "GR" : GR,
}



"""
These are the functions used to process medline (pubmed) files at the backend. They are meant for use internal use by metaknowledge.
"""
from .recordMedline import MedlineRecord, medlineRecordParser
from .medlineHandlers import isMedlineFile, medlineParser
from .tagProcessing.tagNames import tagNameDict, authorBasedTags, tagNameConverterDict
from .tagProcessing.specialFunctions import medlineSpecialTagToFunc
from .tagProcessing.tagFunctions import *



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
monthDict = {'SPR': 3, 'SUM': 6, 'FAL': 9, 'WIN': 12, 'JAN' : 1, 'FEB' : 2, 'MAR' : 3, 'APR' : 4, 'MAY' : 5, 'JUN' : 6 , 'JUL' : 7, 'AUG' : 8, 'SEP' : 9, 'OCT' : 10, 'NOV' : 11, 'DEC' : 12}

def getMonth(s):
    """
    Known formats:
    Month ("%b")
    Month Day ("%b %d")
    Month-Month ("%b-%b") --- this gets coerced to the first %b, dropping the month range
    Season ("%s") --- this gets coerced to use the first month of the given season
    Month Day Year ("%b %d %Y")
    Month Year ("%b %Y")
    Year Month Day ("%Y %m %d")
    """
    monthOrSeason = s.split('-')[0].upper()
    if monthOrSeason in monthDict:
        return monthDict[monthOrSeason]
    else:
        monthOrSeason = s.split('-')[1].upper()
        if monthOrSeason.isdigit():
            return monthOrSeason
        else:
            return monthDict[monthOrSeason]

    raise ValueError("Month format not recognized: " + s)

def makeBiDirectional(d):
    """
    Helper for generating tagNameConverter
    Makes dict that maps from key to value and back
    """
    dTmp = d.copy()
    for k in d:
        dTmp[d[k]] = k
    return dTmp

def reverseDict(d):
    """
    Helper for generating fullToTag
    Makes dict of value to key
    """
    retD = {}
    for k in d:
        retD[d[k]] = k
    return retD
